// The following page is cached from a copy generated at 2021-05-06 05:21:16.
// It is 62 seconds old.
// Proof-of-work difficulty table, normalised such
// that the pentadecathlon has unit difficulty.
// soup count: 279867825847143
difficulties[                                                      "xp2_7"] = 0.0;
difficulties[                                                    "xq4_153"] = 0.0;
difficulties[                                                     "xp2_7e"] = 0.0;
difficulties[                                                   "xp2_318c"] = 0.0;
difficulties[                                                   "xq4_6frc"] = 0.0;
difficulties[              "xp3_co9nas0san9oczgoldlo0oldlogz1047210127401"] = 0.0;
difficulties[                                                 "xq4_27dee6"] = 0.0;
difficulties[                                                "xq4_27deee6"] = 0.0;
difficulties[                                               "xp15_4r4z4r4"] = 1.0;
difficulties[                "yl144_1_16_afb5f3db909e60548f086e22ee3353ac"] = 2.0;
difficulties[                "yl384_1_59_7aeb1999980c43b4945fb7fcdb023326"] = 4.0;
difficulties[                                                   "xp2_2a54"] = 7.0;
difficulties[                                              "xp8_gk2gb3z11"] = 14.0;
difficulties[                                                  "xp2_31ago"] = 19.0;
difficulties[                                             "xp2_0g0k053z32"] = 25.0;
difficulties[                                                 "xp4_37bkic"] = 35.0;
difficulties[                                          "xp2_g8gid1e8z1226"] = 55.0;
difficulties[                                                "xp2_rhewehr"] = 93.0;
difficulties[                                                        "xp2"] = 120.0;
difficulties[                                             "xp4_ssj3744zw3"] = 159.0;
difficulties[                                               "xp8_g3jgz1ut"] = 234.0;
difficulties[                                     "xp30_w33z8kqrqk8zzzx33"] = 347.0;
difficulties[                                     "xp30_w33z8kqrqk8zzzw33"] = 440.0;
difficulties[                                "xp4_8eh5e0e5he8z178a707a871"] = 595.0;
difficulties[                                            "xp5_idiidiz01w1"] = 767.0;
difficulties[                                               "xp14_j9d0d9j"] = 1011.0;
difficulties[                                             "xp3_695qc8zx33"] = 1278.0;
difficulties[                                             "methuselah_25k"] = 1711.0;
difficulties[                                             "methuselah_26k"] = 2495.0;
difficulties[                                            "xp6_ccb7w66z066"] = 3318.0;
difficulties[                                             "methuselah_27k"] = 4119.0;
difficulties[                                           "xp8_wgovnz234z33"] = 5338.0;
difficulties[                                                        "xp3"] = 7055.0;
difficulties[                                             "methuselah_28k"] = 9926.0;
difficulties[                                             "methuselah_29k"] = 14022.0;
difficulties[                                          "xq4_27dee6z4eb776"] = 19774.0;
difficulties[                                                        "xq4"] = 25358.0;
difficulties[                                             "methuselah_30k"] = 33251.0;
difficulties[                                             "methuselah_31k"] = 47423.0;
difficulties[                                            "xp5_3pmwmp3zx11"] = 62074.0;
difficulties[                                             "methuselah_32k"] = 79278.0;
difficulties[                                     "xp15_4r4y14r4z4r4y14r4"] = 100698.0;
difficulties[                               "xp4_8eh5e0e5he8gz178a707a871"] = 124357.0;
difficulties[                                             "methuselah_33k"] = 154922.0;
difficulties[                                                        "xp4"] = 202131.0;
difficulties[                                             "methuselah_34k"] = 243710.0;
difficulties[                         "xp46_330279cx1aad3y833zx4e93x855bc"] = 303318.0;
difficulties[                         "xp46_330279cx1aad3zx4e93x855bcy8cc"] = 374196.0;
difficulties[                                             "methuselah_35k"] = 468431.0;
difficulties[                                                       "xp30"] = 591989.0;
difficulties[                                             "methuselah_36k"] = 747129.0;
difficulties[   "xp15_g0444g0g4440gy14r4z30kkk303kkk03y14r4z70ggg707ggg07"] = 905354.0;
difficulties[                                             "methuselah_37k"] = 1103607.0;
difficulties[                                         "xp15_4r4y0oge2z4r4"] = 1322976.0;
difficulties[                                             "methuselah_38k"] = 1573300.0;
difficulties[                                                     "yl1152"] = 1790693.0;
difficulties[     "xp15_co9nas0san9oczgoldlo0oldlogxuquz1047210127401xfbf"] = 2057997.0;
difficulties[                                               "corderpuffer"] = 2379761.0;
difficulties[                                                       "xp15"] = 2799028.0;
difficulties[                                                        "xp5"] = 3330268.0;
difficulties[                                             "methuselah_39k"] = 3961419.0;
difficulties[                                           "xp8_2erore2z07x7"] = 4737154.0;
difficulties[ "xp24_co9nas0san9oczgoldlo0oldlogy0ggz1047210127401y01581qo"] = 5484673.0;
difficulties[                                             "methuselah_40k"] = 6282306.0;
difficulties[                                        "xq16_gcbgzvgg826frc"] = 7309952.0;
difficulties[                                         "xp8_25du4szx327ba4"] = 8502015.0;
difficulties[   "xp24_co9nas0san9ocy0ssszgoldlo0oldlogy3777z1047210127401"] = 9936090.0;
difficulties[                                             "methuselah_41k"] = 11842886.0;
difficulties[                              "xp46_033y133zzzckgsxsgkczz0cc"] = 14468342.0;
difficulties[                                             "methuselah_42k"] = 17683529.0;
difficulties[                                                        "xp8"] = 21047913.0;
difficulties[                                         "xp6_w8o0uh224a4z32"] = 25919002.0;
difficulties[                                             "methuselah_43k"] = 30751358.0;
difficulties[                                           "xp6_318c0fh884a4"] = 35298252.0;
difficulties[                                         "xp6_wq52z01aq23z32"] = 40498440.0;
difficulties[                                           "xp6_4a422hu0og26"] = 47003371.0;
difficulties[                                                        "xp6"] = 55314943.0;
difficulties[                     "xp46_330279cx1aad3y833zx4e93x855bcyd66"] = 63438117.0;
difficulties[                     "xp46_330279cx1aad3yd66zx4e93x855bcy8cc"] = 73753258.0;
difficulties[                                             "methuselah_44k"] = 88074279.0;
difficulties[                     "xp46_330279cx1aad3yd33zx4e93x855bcy8cc"] = 106725303.0;
difficulties[                     "xp46_330279cx1aad3y833zx4e93x855bcydcc"] = 135397772.0;
difficulties[                                             "methuselah_45k"] = 167993532.0;
difficulties[                                             "methuselah_46k"] = 201592239.0;
difficulties[                                             "methuselah_47k"] = 251990299.0;
difficulties[                                                 "methuselah"] = 335987065.0;
difficulties[                                       "xp16_0c2w3vz33032988"] = 431983370.0;
difficulties[                                          "xq7_3nw17862z6952"] = 604776718.0;
difficulties[                                 "xq12_fh1i0i1hfzw8sms8zxfjf"] = 907165077.0;
difficulties[                                                      "xp120"] = 1511941796.0;
difficulties[                                                        "xp7"] = 3023883592.0;
difficulties[                                                       "yl72"] = 9071650777.0;

