"movups (%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups (%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups (%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, (%3) \n\t"
"movups 16(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 16(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 16(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 16(%3) \n\t"
"movups 32(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 32(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 32(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 32(%3) \n\t"
"movups 48(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 48(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 48(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 48(%3) \n\t"
"movups 64(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 64(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 64(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 64(%3) \n\t"
"movups 80(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 80(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 80(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 80(%3) \n\t"
"movups 96(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 96(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 96(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 96(%3) \n\t"
"movups 112(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 112(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 112(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 112(%3) \n\t"
"movups 128(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 128(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 128(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 128(%3) \n\t"
"movups 144(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 144(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 144(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 144(%3) \n\t"
"movups 160(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 160(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 160(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 160(%3) \n\t"
"movups 176(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 176(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 176(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 176(%3) \n\t"
"movups 192(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 192(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 192(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 192(%3) \n\t"
"movups 208(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 208(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 208(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 208(%3) \n\t"
"movups 224(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 224(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 224(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 224(%3) \n\t"
"movups 240(%0), %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"paddb %%xmm0, %%xmm0 \n\t"
"movups 240(%1), %%xmm1 \n\t"
"paddb %%xmm0, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"paddb %%xmm1, %%xmm1 \n\t"
"movups 240(%2), %%xmm0 \n\t"
"paddb %%xmm1, %%xmm0 \n\t"
"movups %%xmm0, 240(%3) \n\t"
