"vmovups (%0), %%ymm0 \n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vmovups (%1), %%ymm1 \n\t"
"vpaddb %%ymm0, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vmovups (%2), %%ymm0 \n\t"
"vpaddb %%ymm1, %%ymm0, %%ymm0\n\t"
"vmovups %%ymm0, (%3) \n\t"
"vmovups 32(%0), %%ymm0 \n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vmovups 32(%1), %%ymm1 \n\t"
"vpaddb %%ymm0, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vmovups 32(%2), %%ymm0 \n\t"
"vpaddb %%ymm1, %%ymm0, %%ymm0\n\t"
"vmovups %%ymm0, 32(%3) \n\t"
"vmovups 64(%0), %%ymm0 \n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vmovups 64(%1), %%ymm1 \n\t"
"vpaddb %%ymm0, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vmovups 64(%2), %%ymm0 \n\t"
"vpaddb %%ymm1, %%ymm0, %%ymm0\n\t"
"vmovups %%ymm0, 64(%3) \n\t"
"vmovups 96(%0), %%ymm0 \n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vmovups 96(%1), %%ymm1 \n\t"
"vpaddb %%ymm0, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vmovups 96(%2), %%ymm0 \n\t"
"vpaddb %%ymm1, %%ymm0, %%ymm0\n\t"
"vmovups %%ymm0, 96(%3) \n\t"
"vmovups 128(%0), %%ymm0 \n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vmovups 128(%1), %%ymm1 \n\t"
"vpaddb %%ymm0, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vmovups 128(%2), %%ymm0 \n\t"
"vpaddb %%ymm1, %%ymm0, %%ymm0\n\t"
"vmovups %%ymm0, 128(%3) \n\t"
"vmovups 160(%0), %%ymm0 \n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vmovups 160(%1), %%ymm1 \n\t"
"vpaddb %%ymm0, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vmovups 160(%2), %%ymm0 \n\t"
"vpaddb %%ymm1, %%ymm0, %%ymm0\n\t"
"vmovups %%ymm0, 160(%3) \n\t"
"vmovups 192(%0), %%ymm0 \n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vmovups 192(%1), %%ymm1 \n\t"
"vpaddb %%ymm0, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vmovups 192(%2), %%ymm0 \n\t"
"vpaddb %%ymm1, %%ymm0, %%ymm0\n\t"
"vmovups %%ymm0, 192(%3) \n\t"
"vmovups 224(%0), %%ymm0 \n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vpaddb %%ymm0, %%ymm0, %%ymm0\n\t"
"vmovups 224(%1), %%ymm1 \n\t"
"vpaddb %%ymm0, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vpaddb %%ymm1, %%ymm1, %%ymm1\n\t"
"vmovups 224(%2), %%ymm0 \n\t"
"vpaddb %%ymm1, %%ymm0, %%ymm0\n\t"
"vmovups %%ymm0, 224(%3) \n\t"
