"vmovups (%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups (%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups (%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, (%3) \n\t"
"vmovups 16(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 16(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 16(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 16(%3) \n\t"
"vmovups 32(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 32(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 32(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 32(%3) \n\t"
"vmovups 48(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 48(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 48(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 48(%3) \n\t"
"vmovups 64(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 64(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 64(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 64(%3) \n\t"
"vmovups 80(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 80(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 80(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 80(%3) \n\t"
"vmovups 96(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 96(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 96(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 96(%3) \n\t"
"vmovups 112(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 112(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 112(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 112(%3) \n\t"
"vmovups 128(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 128(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 128(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 128(%3) \n\t"
"vmovups 144(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 144(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 144(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 144(%3) \n\t"
"vmovups 160(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 160(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 160(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 160(%3) \n\t"
"vmovups 176(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 176(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 176(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 176(%3) \n\t"
"vmovups 192(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 192(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 192(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 192(%3) \n\t"
"vmovups 208(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 208(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 208(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 208(%3) \n\t"
"vmovups 224(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 224(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 224(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 224(%3) \n\t"
"vmovups 240(%0), %%xmm0 \n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vpaddb %%xmm0, %%xmm0, %%xmm0\n\t"
"vmovups 240(%1), %%xmm1 \n\t"
"vpaddb %%xmm0, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vpaddb %%xmm1, %%xmm1, %%xmm1\n\t"
"vmovups 240(%2), %%xmm0 \n\t"
"vpaddb %%xmm1, %%xmm0, %%xmm0\n\t"
"vmovups %%xmm0, 240(%3) \n\t"
